/* dlgauto.h automaton
 *
 * SOFTWARE RIGHTS
 *
 * We reserve no LEGAL rights to the Purdue Compiler Construction Tool
 * Set (PCCTS) -- PCCTS is in the public domain.  An individual or
 * company may do whatever they wish with source code distributed with
 * PCCTS or the code generated by PCCTS, including the incorporation of
 * PCCTS, or its output, into commerical software.
 * 
 * We encourage users to develop software with PCCTS.  However, we do ask
 * that credit is given to us for developing PCCTS.  By "credit",
 * we mean that if you incorporate our source code into one of your
 * programs (commercial product, research project, or otherwise) that you
 * acknowledge this fact somewhere in the documentation, research report,
 * etc...  If you like PCCTS and have developed a nice tool with the
 * output, please mention that you developed it using PCCTS.  In
 * addition, we ask that this header remain intact in our source code.
 * As long as these guidelines are kept, we expect to continue enhancing
 * this system and expect to make other tools available as they are
 * completed.
 *
 * ANTLR 1.06
 * Will Cohen
 * Purdue University
 * 1989-1992
 */

char	*zzlextext;	/* text of most recently matched token */
char	*zzbegexpr;	/* beginning of last reg expr recogn. */
char	*zzendexpr;	/* beginning of last reg expr recogn. */
int	zzbufsize;	/* number of characters in zzlextext */
int	zzbegcol = 0;	/* column that first character of token is in*/
int	zzendcol = 0;	/* column that last character of token is in */
int	zzline = 1;	/* line current token is on */
int	zzchar;		/* character to determine next state */
int	zzbufovf;	/* indicates that buffer too small for text */
int	zzcharfull = 0;
static char	*zznextpos;/* points to next available position in zzlextext*/
static int 	zzclass;

#ifdef __STDC__
void	zzerrstd(char *);
void	(*zzerr)(char *)=zzerrstd;/* pointer to error reporting function */
#else
void	zzerrstd();
void	(*zzerr)()=zzerrstd;	/* pointer to error reporting function */
#endif

static FILE	*zzstream_in;
extern int	zzerr_in();
static int	(*zzfunc_in)() = zzerr_in;

static int 	zzauto = 0;
static int	zzadd_erase;
static char 	zzebuf[70];

#ifdef ZZCOL
#define ZZINC (++zzendcol)
#else
#define ZZINC
#endif


#define ZZGETC_STREAM {zzchar = getc(zzstream_in); zzclass = ZZSHIFT(zzchar);}

#define ZZGETC_FUNC {zzchar = (*zzfunc_in)(); zzclass = ZZSHIFT(zzchar);}

void
#ifdef __STDC__
zzrdstream( FILE *f )
#else
zzrdstream( f )
FILE *f;
#endif
{
	/* make sure that it is really set to something, otherwise just 
	   leave it be.
	*/
	if (f){
		/* make sure that there is always someplace to get input
		   before closing zzstream_in
		*/
		if (zzstream_in && zzstream_in!=stdin) fclose( zzstream_in );
		zzline = 1;
		zzstream_in = f;
		zzfunc_in = NULL;
		zzcharfull = 0;
	}
}

void
#ifdef __STDC__
zzrdfunc( int (*f)() )
#else
zzrdfunc( f )
int (*f)();
#endif
{
	/* make sure that it is really set to something, otherwise just 
	   leave it be.
	*/
	if (f){
		/* make sure that there is always someplace to get input
		   before closing zzstream_in
		*/
		if (zzstream_in && zzstream_in!=stdin) fclose( zzstream_in );
		zzline = 1;
		zzstream_in = NULL;
		zzfunc_in = f;
		zzcharfull = 0;
	}
}


void
zzclose_stream()
{
#if 0
	fclose( zzstream_in );
	zzstream_in = NULL;
	zzfunc_in = NULL;
#endif
}

void
#ifdef __STDC__
zzmode( int m )
#else
zzmode( m )
int m;
#endif
{
	/* points to base of dfa table */
	if (m<MAX_MODE){
		zzauto = m;
		/* have to redo class since using different compression */
		zzclass = ZZSHIFT(zzchar);
	}else{
		sprintf(zzebuf,"Invalid automaton mode = %d ",m);
		zzerr(zzebuf);
	}
}

/* erase what is currently in the buffer, and get a new reg. expr */
void
zzskip()
{
	zzadd_erase = 1;
}

/* don't erase what is in the zzlextext buffer, add on to it */
void
zzmore()
{
	zzadd_erase = 2;
}

/* substitute c for the reg. expr last matched and is in the buffer */
#ifdef __STDC__
void
zzreplchar(char c)
#else
void
zzreplchar(c)
char c;
#endif
{
	/* can't allow overwriting null at end of string */
	if (zzbegexpr < &zzlextext[zzbufsize-1]){
		*zzbegexpr = c;
		*(zzbegexpr+1) = '\0';
	}
	zzendexpr = zzbegexpr;
	zznextpos = zzbegexpr + 1;
}

/* replace the string s for the reg. expr last matched and in the buffer */
void
#ifdef __STDC__
zzreplstr(register char *s)
#else
zzreplstr(s)
register char *s;
#endif
{
	register char *l= &zzlextext[zzbufsize -1];

	zznextpos = zzbegexpr;
	if (s){
	 	while ((zznextpos <= l) && (*(zznextpos++) = *(s++))){
			/* empty */
		}
		/* correct for NULL at end of string */
		zznextpos--;
	}
	if ((zznextpos <= l) && (*(--s) == 0)){
		zzbufovf = 0;
	}else{
		zzbufovf = 1;
	}
	*(zznextpos) = '\0';
	zzendexpr = zznextpos - 1;
}

void
zzgettok()
{
	register int state, newstate;
	/* last space reserved for the null char */
	register char *lastpos;

skip:
	zzbufovf = 0;
	lastpos = &zzlextext[zzbufsize-1];
	zznextpos = zzlextext;
	zzbegcol = zzendcol+1;
more:
	zzbegexpr = zznextpos;
#ifdef ZZINTERACTIVE
	/* interactive version of automaton */
	/* if there is something in zzchar, process it */
	state = newstate = dfa_base[zzauto];
	if (zzcharfull){
		ZZINC;
		/* Truncate matching buffer to size (not an error) */
		if (zznextpos < lastpos){
			*(zznextpos++) = zzchar;
		}else{
			zzbufovf = 1;
		}
		newstate = dfa[state][zzclass];
	}
	if (zzstream_in)
		while (zzalternatives[newstate]){
			state = newstate;
			ZZGETC_STREAM;
			ZZINC;
			/* Truncate matching buffer to size (not an error) */
			if (zznextpos < lastpos){
				*(zznextpos++) = zzchar;
			}else{
				zzbufovf = 1;
			}
			newstate = dfa[state][zzclass];
		}
	else if (zzfunc_in)
		while (zzalternatives[newstate]){
			state = newstate;
			ZZGETC_STREAM;
			ZZINC;
			/* Truncate matching buffer to size (not an error) */
			if (zznextpos < lastpos){
				*(zznextpos++) = zzchar;
			}else{
				zzbufovf = 1;
			}
			newstate = dfa[state][zzclass];
		}
	/* figure out if last character really part of token */
	if ((state != dfa_base[zzauto]) && (newstate == DfaStates)){
		zzcharfull = 1;
		--zznextpos;
	}else{
		zzcharfull = 0;
		state = newstate;
	}
	*(zznextpos) = '\0';
	/* Able to transition out of start state to some non err state?*/
	if ( state == dfa_base[zzauto] ){
		/* make sure doesn't get stuck */
		zzadvance();
	}
#else
	/* non-interactive version of automaton */
	if (!zzcharfull)
		zzadvance();
	else
		ZZINC;
	state = dfa_base[zzauto];
	if (zzstream_in)
		while ((newstate = dfa[state][zzclass]) != DfaStates){
			state = newstate;
			/* Truncate matching buffer to size (not an error) */
			if (zznextpos < lastpos){
				*(zznextpos++) = zzchar;
			}else{
				zzbufovf = 1;
			}
			ZZGETC_STREAM;
			ZZINC;
		}
	else if (zzfunc_in)
		while ((newstate = dfa[state][zzclass]) != DfaStates){
			state = newstate;
			/* Truncate matching buffer to size (not an error) */
			if (zznextpos < lastpos){
				*(zznextpos++) = zzchar;
			}else{
				zzbufovf = 1;
			}
			ZZGETC_FUNC;
			ZZINC;
		}
	zzcharfull = 1;
	if ( state == dfa_base[zzauto] ){
		if (zznextpos < lastpos){
			*(zznextpos++) = zzchar;
		}else{
			zzbufovf = 1;
		}
		*zznextpos = '\0';
		/* make sure doesn't get stuck */
		zzadvance();
	}else{
		*zznextpos = '\0';
	}
#endif
#ifdef ZZCOL
	zzendcol -= zzcharfull;
#endif
	zzendexpr = zznextpos -1;
	zzadd_erase = 0;
	(*actions[accepts[state]])();
	switch (zzadd_erase) {
		case 1: goto skip;
		case 2: goto more;
	}
}

void
zzadvance()
{
	if (zzstream_in) { ZZGETC_STREAM; zzcharfull = 1; ZZINC;}
	if (zzfunc_in) { ZZGETC_FUNC; zzcharfull = 1; ZZINC;}
	if (!(zzstream_in || zzfunc_in)){
		zzerr_in();
	}
}

void
#ifdef __STDC__
zzerrstd(char *s)
#else
zzerrstd(s)
char *s;
#endif
{
        fprintf(stderr,
                "%s near line %d (text was '%s')\n",
                ((s == NULL) ? "Lexical error" : s),
                zzline,zzlextext);
}

int
zzerr_in()
{
	fprintf(stderr,"No input stream or function\n");
	/* return eof to get out gracefully */
	return -1;
}
